"""Type aliases and generic type variables for the dictIO package."""

# ruff: noqa: ERA001

from collections.abc import Hashable, MutableMapping, MutableSequence
from typing import Any, TypeAlias, TypeVar

# Type aliases for keys
TKey: TypeAlias = Hashable
TGlobalKey: TypeAlias = TKey | int

# Type aliases for values
TValue: TypeAlias = Any
TSingleValue: TypeAlias = str | int | float | bool | None

# Generic Type Variables
K = TypeVar("K", bound=TKey)
V = TypeVar("V", bound=TValue)
M = TypeVar("M", bound=MutableMapping[K, V])  # type: ignore[valid-type, reportGeneralTypeIssues]
S = TypeVar("S", bound=MutableSequence[V])  # type: ignore[valid-type, reportGeneralTypeIssues]



# pyright: reportPrivateUsage=false
# pyright: reportUnnecessaryTypeIgnoreComment=false

import os
import re
import sys
from copy import deepcopy
from functools import partial
from pathlib import Path, PurePath

import pytest
from numpy.testing import assert_array_equal

from dictIO import DictReader, DictWriter, NativeParser, SDict
from dictIO.types import TKey, TValue

WindowsOnly: pytest.MarkDecorator = pytest.mark.skipif(not sys.platform.startswith("win"), reason="windows only test")


def test_file_not_found_exception() -> None:
    # Prepare
    source_file = Path("this_file_does_not_exist")
    # Execute and Assert
    with pytest.raises(FileNotFoundError):
        _ = DictReader.read(source_file)


def test_merge_includes() -> None:
    # sourcery skip: class-extract-method
    # Prepare dict until and including _parse_tokenized_dict()
    s_dict: SDict[TKey, TValue] = SDict()
    SetupHelper.prepare_dict_until(s_dict, until_step=-1)
    dict_in = deepcopy(s_dict)
    # Assert dict_in
    assert dict_in["expressions"]["reference"]["value"][:10] == "EXPRESSION"
    assert dict_in["expressions"]["expression1"]["value"][:10] == "EXPRESSION"
    assert dict_in["expressions"]["expression2"]["value"][:10] == "EXPRESSION"
    assert dict_in["expressions"]["expression3"]["value"][:10] == "EXPRESSION"
    # Execute
    DictReader._merge_includes(s_dict)
    dict_out = s_dict
    # Assert
    assert len(dict_out) == len(dict_in) + 8
    assert dict_out["paramA"] == 3.0
    assert dict_out["paramB"] == 4.0
    assert dict_out["paramC"] == 7.0
    assert dict_out["paramD"] == 0.66
    assert dict_out["paramE"] == [0.1, 0.2, 0.4]
    assert dict_out["paramF"] == [[0.3, 0.9], [2.7, 8.1]]
    assert dict_out["paramG"] == [[10, "fancy", 3.14, "s"], ["more", 2, "come"]]


def test_resolve_reference() -> None:
    # Prepare dict until and including ()
    s_dict: SDict[TKey, TValue] = SDict()
    SetupHelper.prepare_dict_until(s_dict, until_step=0)
    # Assert non-indexed references have been resolved
    assert DictReader._resolve_reference("$paramA", s_dict.variables) == 3.0
    assert DictReader._resolve_reference("$paramB", s_dict.variables) == 4.0
    assert DictReader._resolve_reference("$paramC", s_dict.variables) == 7.0
    assert DictReader._resolve_reference("$paramD", s_dict.variables) == 0.66
    assert DictReader._resolve_reference("$paramE[0]", s_dict.variables) == 0.1

    paramE = DictReader._resolve_reference("$paramE", s_dict.variables)  # noqa: N806
    assert isinstance(paramE, list)
    assert len(paramE) == 3
    assert paramE[0] == 0.1
    assert paramE[1] == 0.2
    assert paramE[2] == 0.4
    paramF = DictReader._resolve_reference("$paramF", s_dict.variables)  # noqa: N806
    assert isinstance(paramF, list)
    assert len(paramF) == 2
    assert len(paramF[0]) == 2
    assert len(paramF[1]) == 2
    assert paramF[0][0] == 0.3
    assert paramF[0][1] == 0.9
    assert paramF[1][0] == 2.7
    assert paramF[1][1] == 8.1
    paramG = DictReader._resolve_reference("$paramG", s_dict.variables)  # noqa: N806
    assert isinstance(paramG, list)
    assert len(paramG) == 2
    assert len(paramG[0]) == 4
    assert len(paramG[1]) == 3
    assert paramG[0][0] == 10
    assert paramG[0][1] == "fancy"
    assert paramG[0][2] == 3.14
    assert paramG[0][3] == "s"
    assert paramG[1][0] == "more"
    assert paramG[1][1] == 2
    assert paramG[1][2] == "come"

    # Assert indexed references have been resolved
    assert DictReader._resolve_reference("$paramE[0]", s_dict.variables) == 0.1
    assert DictReader._resolve_reference("$paramE[1]", s_dict.variables) == 0.2
    assert DictReader._resolve_reference("$paramE[2]", s_dict.variables) == 0.4
    assert DictReader._resolve_reference("$paramF[0][0]", s_dict.variables) == 0.3
    assert DictReader._resolve_reference("$paramF[0][1]", s_dict.variables) == 0.9
    assert DictReader._resolve_reference("$paramF[1][0]", s_dict.variables) == 2.7
    assert DictReader._resolve_reference("$paramF[1][1]", s_dict.variables) == 8.1
    assert DictReader._resolve_reference("$paramG[0][0]", s_dict.variables) == 10
    assert DictReader._resolve_reference("$paramG[0][1]", s_dict.variables) == "fancy"
    assert DictReader._resolve_reference("$paramG[0][2]", s_dict.variables) == 3.14
    assert DictReader._resolve_reference("$paramG[0][3]", s_dict.variables) == "s"
    assert DictReader._resolve_reference("$paramG[1][0]", s_dict.variables) == "more"
    assert DictReader._resolve_reference("$paramG[1][1]", s_dict.variables) == 2
    assert DictReader._resolve_reference("$paramG[1][2]", s_dict.variables) == "come"


def test_eval_expressions() -> None:
    # Prepare dict until and including ()
    s_dict: SDict[TKey, TValue] = SDict()
    SetupHelper.prepare_dict_until(s_dict, until_step=0)
    dict_in = deepcopy(s_dict)
    # Assert dict_in
    assert dict_in["expressions"]["reference"]["value"][:10] == "EXPRESSION"  # $paramA
    assert dict_in["expressions"]["expression1"]["value"][:10] == "EXPRESSION"  # "$paramB"
    assert dict_in["expressions"]["expression2"]["value"][:10] == "EXPRESSION"  # "$paramC + 4"
    assert dict_in["expressions"]["expression3"]["value"][:10] == "EXPRESSION"  # "$paramC + $paramD"
    assert dict_in["expressions"]["expressionE"]["value"][:10] == "EXPRESSION"  # $paramE[0]
    assert dict_in["expressions"]["expressionF"]["value"][:10] == "EXPRESSION"  # $paramF[0][0]
    assert dict_in["expressions"]["expressionG1"]["value"][:10] == "EXPRESSION"  # "$paramG"
    assert dict_in["expressions"]["expressionG2"]["value"][:10] == "EXPRESSION"  # "$paramG[0]"
    assert dict_in["expressions"]["expressionG3"]["value"][:10] == "EXPRESSION"  # "$paramG[1][2]"

    # Execute
    DictReader._eval_expressions(s_dict)
    dict_out = s_dict

    # Assert references have been resolved
    assert dict_out["expressions"]["reference"]["value"] == 3.0  # 3.0
    assert dict_out["expressions"]["expression1"]["value"] == 4.0  # 4.0
    assert dict_out["expressions"]["expression2"]["value"] == 11.0  # 7.0 + 4
    assert dict_out["expressions"]["expression3"]["value"] == 7.66  # 7.0 + 0.66
    assert dict_out["expressions"]["expressionE"]["value"] == 0.1  # paramE[0]
    assert dict_out["expressions"]["expressionF"]["value"] == 0.3  # paramF[0][0]
    assert dict_out["expressions"]["expressionG1"]["value"] == [
        [10, "fancy", 3.14, "s"],
        ["more", 2, "come"],
    ]  # paramG
    assert dict_out["expressions"]["expressionG2"]["value"] == [
        10,
        "fancy",
        3.14,
        "s",
    ]  # paramG[0]
    assert dict_out["expressions"]["expressionG3"]["value"] == "come"  # paramG[1][2]


def test_eval_expressions_with_included_keys() -> None:
    # test keys with the same name as included keys
    # Prepare
    source_file = Path("test_dictReader_dict")

    # Execute
    dict_in = DictReader.read(source_file, includes=True)
    dict_out = dict_in

    # Assert root keys
    assert dict_out["keyA"] == 3.0  # $paramA
    assert dict_out["keyB"] == 4.0  # "$paramB"
    assert dict_out["keyC"] == 7.0  # "$paramC"
    assert dict_out["keyD"] == 4.16  # "$paramD+$paramC/2"
    assert dict_out["keyE"] == 6.67  # "$paramC-$paramD/2"
    assert dict_out["keyF"] == 8.98  # "3 * $paramD + $paramC"
    assert dict_out["keyG"] == 20.34  # "3 * $paramC - $paramD"
    assert dict_out["keyH"] == 0.2  # "$paramE[1]"
    assert dict_out["keyI"] == [0.1, 0.2, 0.4]  # "$paramE"
    assert dict_out["keyJ"] == 8.1  # "$paramF[1][1]"
    assert dict_out["keyK"] == [2.7, 8.1]  # "$paramF[1]"
    assert dict_out["keyL"] == [[0.3, 0.9], [2.7, 8.1]]  # "$paramF"
    assert dict_out["keyM"] == 9.0  # "3 * $paramA"
    assert dict_out["keyN"] == 7.0  # "$paramA + $paramB"

    # Assert different key names
    assert dict_out["differentKeyNames"]["keyA"] == 3.0  # $paramA
    assert dict_out["differentKeyNames"]["keyB"] == 4.0  # "$paramB"
    assert dict_out["differentKeyNames"]["keyC"] == 7.0  # "$paramC"
    assert dict_out["differentKeyNames"]["keyD"] == 4.16  # "$paramD+$paramC/2"
    assert dict_out["differentKeyNames"]["keyE"] == 6.67  # "$paramC-$paramD/2"
    assert dict_out["differentKeyNames"]["keyF"] == 8.98  # "3 * $paramD + $paramC"
    assert dict_out["differentKeyNames"]["keyG"] == 20.34  # "3 * $paramC - $paramD"
    assert dict_out["differentKeyNames"]["keyH"] == 0.2  # "$paramE[1]"
    assert dict_out["differentKeyNames"]["keyI"] == [0.1, 0.2, 0.4]  # "$paramE"
    assert dict_out["differentKeyNames"]["keyJ"] == 8.1  # "$paramF[1][1]"
    assert dict_out["differentKeyNames"]["keyK"] == [2.7, 8.1]  # "$paramF[1]"
    assert dict_out["differentKeyNames"]["keyL"] == [
        [0.3, 0.9],
        [2.7, 8.1],
    ]  # "$paramF"
    assert dict_out["differentKeyNames"]["keyM"] == 9.0  # "3 * $paramA"
    assert dict_out["differentKeyNames"]["keyN"] == 7.0  # "$paramA + $paramB"

    # Assert same key name as reference name
    # Be aware that in the nested dict 'sameKeyNames', paramD becomes reaasigned.
    # This overwrites paramD from included dict due to the flat lookup table provided through dict.variables
    assert dict_out["sameKeyNames"]["paramA"] == 3.0  # $paramA;
    assert dict_out["sameKeyNames"]["paramB"] == 4.0  # "$paramB";
    assert dict_out["sameKeyNames"]["paramC"] == 7.0  # "$paramC";
    assert dict_out["sameKeyNames"]["paramD"] == 10.5  # "$paramC+$paramC/2";
    assert dict_out["sameKeyNames"]["paramE"] == 0.2  # "$paramE[1]";
    assert dict_out["sameKeyNames"]["paramF"] == [[0.3, 0.9], [2.7, 8.1]]  # "$paramF";

    # Assert references to variables in deeper structure of the included dict
    assert dict_out["keysWithNestedRefs"]["nestKeyA"] == 3.0  # $paramA;
    assert dict_out["keysWithNestedRefs"]["nestKeyB"] == 4.0  # "$paramB";
    assert dict_out["keysWithNestedRefs"]["nestKeyC"] == 12.0  # "$paramA * $paramB";
    assert dict_out["keysWithNestedRefs"]["nestKeyD"] == 35.4  # "$paramC / $paramE[1] + $paramE[2]";
    assert dict_out["keysWithNestedRefs"]["nestKeyE"] == 0.4  # "$paramE[2]";
    assert dict_out["keysWithNestedRefs"]["nestKeyF"] == [
        [0.3, 0.9],
        [2.7, 8.1],
    ]  # "$paramF";
    assert dict_out["keysWithNestedRefs"]["nestKeyH"] == 1.2  # "$paramH";
    assert dict_out["keysWithNestedRefs"]["nestKeyI"] == 3.4  # "$paramI[1]";
    assert dict_out["keysWithNestedRefs"]["nestKeyJ"] == 5.6  # "$paramJ[0][1]";
    assert dict_out["keysWithNestedRefs"]["nestParamA"] == 3.0  # "$paramA";
    assert dict_out["keysWithNestedRefs"]["nestParamB"] == 4.0  # "$paramB";
    assert dict_out["keysWithNestedRefs"]["nestParamC"] == 7.0  # "$paramC";
    assert dict_out["keysWithNestedRefs"]["nestParamD"] == 0.66  # "$paramD";
    assert dict_out["keysWithNestedRefs"]["nestParamE"] == [0.1, 0.2, 0.4]  # "$paramE";
    assert dict_out["keysWithNestedRefs"]["nestParamF"] == 0.3  # "$paramF[0][0]";
    assert dict_out["keysWithNestedRefs"]["nestParamH"] == 1.2  # "$paramH";
    assert dict_out["keysWithNestedRefs"]["nestParamI"] == [2.3, 3.4]  # "$paramI";
    assert dict_out["keysWithNestedRefs"]["nestParamJ"] == [
        [4.5, 5.6],
        [6.7, 7.8],
    ]  # "$paramJ";
    assert dict_out["keysWithNestedRefs"]["nestParamK"] == 0.4  # "$nestKeyE" == "$paramE[2]";
    assert dict_out["keysWithNestedRefs"]["nestParamL"] == 7.0  # "$paramE[2] * 10 + $paramA";
    assert dict_out["keysWithNestedRefs"]["nestParamM"] == 14.8  # "$nestParamJ[1][1] + $paramC";

    # Assert keys that do not point to a single expression, but a list of expressions
    assert dict_out["keysPointingToAListOfExpressions"]["keyToListA"][0] == 3.0  # $paramA;
    assert dict_out["keysPointingToAListOfExpressions"]["keyToListA"][1] == 1
    assert dict_out["keysPointingToAListOfExpressions"]["keyToListA"][2] == 2
    assert dict_out["keysPointingToAListOfExpressions"]["keyToListB"][0] == 4.0  # "$paramB";
    assert dict_out["keysPointingToAListOfExpressions"]["keyToListL"][0] == 14.8  # "$nestParamJ[1][1] + $paramC";


def test_eval_expressions_with_included_numpy_expressions() -> None:
    # Prepare
    source_file = Path("test_dictReader_dict")

    # Execute
    dict_in = DictReader.read(source_file, includes=True)
    dict_out = dict_in
    assert dict_out["keysContainingNumpyExpressions"]["npKeyA"] == 2

    assert_array_equal(
        dict_out["keysContainingNumpyExpressions"]["npKeyB"],
        [
            [1, 1],
            [1, 1],
        ],  # pyright: ignore[reportArgumentType]
    )

    assert_array_equal(
        dict_out["keysContainingNumpyExpressions"]["npKeyC"],
        [2, 2, 2],
    )

    assert_array_equal(
        dict_out["keysContainingNumpyExpressions"]["npKeyD"],
        [
            [2, 0, 0],
            [0, 2, 0],
            [0, 0, 2],
        ],  # pyright: ignore[reportArgumentType]
    )

    assert_array_equal(
        dict_out["keysContainingNumpyExpressions"]["npKeyE"],
        [
            [2, 0, 0],
            [0, 2, 0],
            [0, 0, 2],
        ],  # pyright: ignore[reportArgumentType]
    )

    assert_array_equal(
        dict_out["keysContainingNumpyExpressions"]["npKeyZ"],
        [
            [0, 0, 0, 0],
        ],  # pyright: ignore[reportArgumentType]
    )


def test_reread_string_literals() -> None:
    # Prepare
    source_file = Path("test_dictReader_dict")
    parsed_file = Path(f"parsed.{source_file.name}")
    parsed_file.unlink(missing_ok=True)
    dict_in = DictReader.read(source_file, includes=True)
    DictWriter.write(dict_in, parsed_file)
    assert parsed_file.exists()
    # Execute
    reread_dict = DictReader.read(parsed_file)
    # Asset that string literals containing the '$' character are written back
    # and reread with surrounding single quotes.
    # This to avoid that a string literal with single quotes that contains a '$' character
    # gets unintentionally evaluated as expression when rereading a parsed dict.
    assert dict_in["differentKeyNames"] == reread_dict["differentKeyNames"]
    assert dict_in["sameKeyNames"] == reread_dict["sameKeyNames"]
    # Clean up
    parsed_file.unlink()


# @TODO: To be implemented
@pytest.mark.skip(reason="To be implemented")
def test_remove_comment_keys() -> None:
    pass


# @TODO: To be implemented
@pytest.mark.skip(reason="To be implemented")
def test_remove_include_keys() -> None:
    pass


def test_read_dict() -> None:
    # Prepare
    source_file = Path("test_dictReader_dict")
    # Execute
    dict_in = DictReader.read(source_file)
    # Assert included dict has been merged
    assert dict_in["paramA"] == 3.0
    assert dict_in["paramB"] == 4.0
    assert dict_in["paramC"] == 7.0
    assert dict_in["paramD"] == 0.66
    assert dict_in["paramE"] == [0.1, 0.2, 0.4]
    assert dict_in["paramF"] == [[0.3, 0.9], [2.7, 8.1]]
    assert dict_in["paramG"] == [[10, "fancy", 3.14, "s"], ["more", 2, "come"]]
    # Assert references and expressions have been resolved
    assert dict_in["keyA"] == 3.0
    assert dict_in["keyB"] == 4.0
    assert dict_in["keyC"] == 7.0
    assert dict_in["keyD"] == 4.16
    assert dict_in["keyE"] == 6.67
    assert dict_in["keyF"] == 8.98
    assert dict_in["keyG"] == 20.34
    assert dict_in["keyH"] == 0.2
    assert dict_in["keyI"] == [0.1, 0.2, 0.4]
    assert dict_in["keyJ"] == 8.1
    assert dict_in["keyK"] == [2.7, 8.1]
    assert dict_in["keyL"] == [[0.3, 0.9], [2.7, 8.1]]
    assert dict_in["keyM"] == 9.0
    assert dict_in["keyN"] == 7.0


def test_read_json() -> None:
    # Prepare
    source_file = Path("test_dictReader_json.json")
    # Execute
    dict_in = DictReader.read(source_file)
    # Assert included dict has been merged
    assert dict_in["paramA"] == 3.0
    assert dict_in["paramB"] == 4.0
    assert dict_in["paramC"] == 7.0
    assert dict_in["paramD"] == 0.66
    assert dict_in["paramE"] == [0.1, 0.2, 0.4]
    assert dict_in["paramF"] == [[0.3, 0.9], [2.7, 8.1]]
    assert dict_in["paramG"] == [[10, "fancy", 3.14, "s"], ["more", 2, "come"]]
    # Assert references and expressions have been resolved
    assert dict_in["keyA"] == 3.0
    assert dict_in["keyB"] == 4.0
    assert dict_in["keyC"] == 7.0
    assert dict_in["keyD"] == 4.16
    assert dict_in["keyE"] == 6.67
    assert dict_in["keyF"] == 8.98
    assert dict_in["keyG"] == 20.34
    assert dict_in["keyH"] == 0.2
    assert dict_in["keyI"] == [0.1, 0.2, 0.4]
    assert dict_in["keyJ"] == 8.1
    assert dict_in["keyK"] == [2.7, 8.1]
    assert dict_in["keyL"] == [[0.3, 0.9], [2.7, 8.1]]
    assert dict_in["keyM"] == 9.0
    assert dict_in["keyN"] == 7.0


def test_compare_expressions_in_dict_format_with_expressions_in_json_format() -> None:
    # sourcery skip: no-loop-in-tests
    # Prepare
    source_file_dict = Path("test_dictReader_dict")
    source_file_json = Path("test_dictReader_json.json")
    # Execute
    dict_dict = DictReader.read(source_file_dict)
    dict_json = DictReader.read(source_file_json)
    # Assert
    assert dict_dict.expressions == dict_json.expressions
    # Collect all references contained in expressions
    references_dict: list[str] = _get_references_in_expressions(dict_dict)
    references_json: list[str] = _get_references_in_expressions(dict_json)
    assert references_dict == references_json
    references_resolved_dict = _resolve_references(dict_dict, references_dict)
    references_resolved_json = _resolve_references(dict_json, references_json)
    assert references_resolved_dict == references_resolved_json
    # assert dict_dict.variables == dict_json.variables
    for dvs, jvs in zip(dict_dict.variables, dict_json.variables, strict=False):
        assert_array_equal(dvs, jvs)


def _get_references_in_expressions(
    s_dict: SDict[TKey, TValue],
) -> list[str]:
    references: list[str] = []
    for item in s_dict.expressions.values():
        _refs: list[str] = re.findall(r"\$\w[\w\[\]]*", item["expression"])
        references.extend(_refs)
    return references


def _resolve_references(
    s_dict: SDict[TKey, TValue],
    references: list[str],
) -> dict[str, TValue | None]:
    # Resolve references
    variables: dict[str, TValue] = s_dict.variables
    references_resolved = {ref: DictReader._resolve_reference(ref, variables) for ref in references}

    return {
        ref: value
        for ref, value in references_resolved.items()
        if (value is not None) and (not re.search(r"EXPRESSION|\$", str(value)))
    }


def test_read_foam() -> None:
    # Prepare
    source_file = Path("test_dictReader_foam")
    # Execute
    dict_in = DictReader.read(source_file)
    # Assert
    assert dict_in["Re_profile"] == 1000000.0
    assert dict_in["mag_U_infty"] == 1.004


def test_read_xml() -> None:
    # Prepare
    source_file = Path("test_dictReader_xml.xml")
    # Execute
    dict_in = DictReader.read(source_file)
    # Assert
    assert "_xmlOpts" in dict_in
    assert len(dict_in["_xmlOpts"]["_nameSpaces"]) == 1
    assert dict_in["_xmlOpts"]["_nameSpaces"]["xs"] == "http://www.w3.org/2001/XMLSchema"
    assert len(dict_in["_xmlOpts"]["_rootAttributes"]) == 1
    assert dict_in["_xmlOpts"]["_rootTag"] == "ROOT"
    assert dict_in["_xmlOpts"]["_rootAttributes"]["version"] == "0.1"
    assert dict_in["_xmlOpts"]["_addNodeNumbering"] is True


@WindowsOnly
def test_read_dict_in_subfolder_bsl() -> None:
    # Prepare
    source_file = Path.cwd() / "subfolder" / "test_subfolder_dict_bsl"
    # Execute
    dict_in = DictReader.read(source_file)
    # Assert
    assert dict_in["valueA"] == 1  # paramA is defined in paramDict_bsl
    assert dict_in["valueB"] == "$paramB"  # paramB is defined in paramDict_fsl


def test_read_dict_in_subfolder_fsl() -> None:
    # Prepare
    source_file = Path.cwd() / "subfolder" / "test_subfolder_dict_fsl"
    # Execute
    dict_in = DictReader.read(source_file)
    # Assert
    assert dict_in["valueA"] == "$paramA"  # paramA is defined in paramDict_bsl
    assert dict_in["valueB"] == 2  # paramB is defined in paramDict_fsl


def test_read_dict_in_subfolder_source_file_given_as_str() -> None:
    # Prepare
    source_file = Path.cwd().absolute() / "subfolder" / "test_subfolder_dict_fsl"
    source_file_as_str = str(source_file)
    # Execute
    dict_in = DictReader.read(source_file_as_str)
    # Assert
    assert dict_in["valueB"] == 2


def test_read_dict_in_subfolder_source_file_given_as_purepath() -> None:
    # Prepare
    source_file = Path.cwd().absolute() / "subfolder" / "test_subfolder_dict_fsl"
    source_file_as_purepath = PurePath(str(source_file))
    # Execute
    dict_in = DictReader.read(source_file_as_purepath)
    # Assert
    assert dict_in["valueB"] == 2  # paramB is defined in paramDict_fsl


def test_read_dict_in_subfolder_parsed_via_dictparser_cli() -> None:
    # Prepare
    source_file = "subfolder/test_subfolder_dict_fsl"
    parsed_file = Path("subfolder/parsed.test_subfolder_dict_fsl")
    parsed_file_foam = Path("subfolder/parsed.test_subfolder_dict_fsl.foam")
    parsed_file.unlink(missing_ok=True)
    parsed_file_foam.unlink(missing_ok=True)
    # Execute
    _ = os.system(f"dictParser --quiet {source_file}")  # noqa: S605
    # Assert
    assert parsed_file.exists()
    assert not parsed_file_foam.exists()
    # Clean up
    parsed_file.unlink()
    # Execute
    _ = os.system(f"dictParser --quiet --output=foam {source_file}")  # noqa: S605
    # Assert
    assert not parsed_file.exists()
    assert parsed_file_foam.exists()
    # Clean up
    parsed_file_foam.unlink()


def test_read_circular_includes() -> None:
    # Prepare
    source_file = Path("circular_include/test_base_dict")
    # Execute
    dict_in = DictReader.read(source_file)
    # Assert
    assert dict_in["baseSubDict"]["baseVar1"] == 2
    assert dict_in["baseSubDict"]["baseVar2"] == 2
    assert dict_in["baseSubDict"]["baseVar3"] == 4
    assert dict_in["baseSubDict"]["baseVar4"] == 8
    assert dict_in["baseSubDict"]["baseVar5"] == 8


def test_read_circular_includes_log_warning(caplog: pytest.LogCaptureFixture) -> None:
    # Prepare
    source_file = Path("circular_include/test_base_dict")
    log_level_expected = "WARNING"
    log_message_expected = (
        "Recursive include detected. "
        "Merging of test_ref1_dict->test_ref2_dict->test_base_dict->test_ref1_dict "
        "into test_base_dict aborted."
    )
    # Execute
    _ = DictReader.read(source_file)
    # Assert
    assert len(caplog.records) == 1
    assert caplog.records[0].levelname == log_level_expected
    assert caplog.records[0].message == log_message_expected


def test_read_strings_dict() -> None:
    # Prepare
    source_file = Path("test_strings_dict")
    # Execute
    dict_in = DictReader.read(source_file)
    # Assert strings are parsed correctly
    assert dict_in["subDict"]["string_00_dq_empty"] == ""
    assert dict_in["subDict"]["string_01_sq_empty"] == ""
    assert dict_in["subDict"]["string_02_dq_word"] == "string_02_dq_word"
    assert dict_in["subDict"]["string_03_sq_word"] == "string_03_sq_word"
    assert dict_in["subDict"]["string_04_dq_sq_word"] == r"quote('string_04_dq_sq_word')"
    assert dict_in["subDict"]["string_05_dq_escsq_word"] == r"quote(\'string_05_dq_escsq_word\')"
    assert dict_in["subDict"]["string_06_sq_dq_word"] == r'quote("string_06_sq_dq_word")'
    assert dict_in["subDict"]["string_07_sq_escdq_word"] == r"quote(\"string_07_sq_escdq_word\")"


def test_reread_strings_dict() -> None:
    # Prepare
    source_file = Path("test_strings_dict")
    parsed_file = Path(f"parsed.{source_file.name}")
    parsed_file.unlink(missing_ok=True)
    # Execute
    dict_in = DictReader.read(source_file)
    DictWriter.write(dict_in, parsed_file)
    assert parsed_file.exists()
    # Assert strings are parsed correctly
    reread_dict = DictReader.read(parsed_file)
    assert reread_dict["subDict"]["string_00_dq_empty"] == ""
    assert reread_dict["subDict"]["string_01_sq_empty"] == ""
    assert reread_dict["subDict"]["string_02_dq_word"] == "string_02_dq_word"
    assert reread_dict["subDict"]["string_03_sq_word"] == "string_03_sq_word"
    assert reread_dict["subDict"]["string_04_dq_sq_word"] == r"quote('string_04_dq_sq_word')"
    assert reread_dict["subDict"]["string_05_dq_escsq_word"] == r"quote(\'string_05_dq_escsq_word\')"
    assert reread_dict["subDict"]["string_06_sq_dq_word"] == r'quote("string_06_sq_dq_word")'
    assert reread_dict["subDict"]["string_07_sq_escdq_word"] == r"quote(\"string_07_sq_escdq_word\")"


def test_single_character_vars() -> None:
    # Prepare
    source_file = Path("test_single_character_vars_dict")
    # Execute
    dict_in = DictReader.read(source_file, includes=False)
    # Assert single character variables are properly parsed
    assert dict_in["a"] == 1.0
    assert dict_in["b"] is True
    assert dict_in["c"]["d"] == 4.0
    assert dict_in["c"]["e"] is False


def test_single_character_references() -> None:
    # Prepare
    source_file = Path("test_single_character_vars_dict")
    # Execute
    dict_in = DictReader.read(source_file)
    # Assert included dict has been merged and single character references been resolved
    assert dict_in["c"]["f"] == 6.0
    assert dict_in["c"]["g"] is True
    assert dict_in["c"]["h"] == 8.0
    assert dict_in["c"]["i"] is False
    assert dict_in["j"] == 10.0
    assert dict_in["k"] == 12.0
    assert dict_in["u"] == 3
    assert dict_in["v"] == "Alice"
    assert dict_in["w"] == "paragliding contest"
    assert dict_in["ww"] == "AliceandBobfailtheparagliding contest"


class SetupHelper:
    @staticmethod
    def prepare_dict_until(
        dict_to_prepare: SDict[TKey, TValue],
        until_step: int = -1,
        file_to_read: str = "test_dictReader_dict",
    ) -> None:
        file_name = Path.cwd() / file_to_read

        parser = NativeParser()
        _ = parser.parse_file(file_name, dict_to_prepare)

        funcs = [
            partial(
                DictReader._merge_includes,  # Step 00
                dict_to_prepare,
            ),
            partial(
                DictReader._eval_expressions,  # Step 01
                dict_to_prepare,
            ),
        ]

        for i in range(until_step + 1):
            funcs[i]()
        return



"""DictWriter class for writing dictionaries in dictIO native file format, as well as JSON and XML."""

import logging
import os
import re
from collections.abc import MutableMapping, MutableSequence
from pathlib import Path

from dictIO import Formatter, NativeParser, SDict, order_keys
from dictIO.types import TKey, TValue

__ALL__ = ["DictWriter", "create_target_file_name"]

logger = logging.getLogger(__name__)


class DictWriter:
    """Writer for dictionaries in dictIO native file format, as well as JSON, XML and OpenFoam."""

    def __init__(self) -> None:
        return

    @staticmethod
    def write(
        source_dict: MutableMapping[TKey, TValue],
        target_file: str | os.PathLike[str] | None = None,
        mode: str = "a",
        *,
        order: bool = False,
        formatter: Formatter | None = None,
    ) -> None:
        """Write a dictionary file in dictIO native file format, as well as JSON, XML and OpenFoam.

        Writes a dictIO dict (parameter source_dict of type SDict) to target_file.
        Following file formats are supported and interpreted through target_file's file ending:
        no file ending   ->   dictIO native dict file
        '.cpp'           ->   dictIO native dict file
        '.foam'          ->   Foam dictionary file
        '.json'          ->   Json dictionary file
        '.xml'           ->   XML file
        Following modes are supported:
        mode = 'a': append to target file. If the existing file contains a dictionary, write() will append the new dict
        to the existing through merging. This is the default behaviour.
        mode = 'w': overwrite target file. The existing file will be overwritten.

        Parameters
        ----------
        source_dict : Union[MutableMapping[TKey, TValue], SDict]
            source dict
        target_file : Union[str, os.PathLike[str], None], optional
            target dict file name, by default None
        mode : str, optional
            append to target file ('a') or overwrite target file ('w'), by default 'a'
        order : bool, optional
            if True, the dict will be sorted before writing, by default False
        formatter : Union[Formatter, None], optional
            formatter to be used, by default None
        """
        # Check arguments
        if mode not in ["a", "w"]:
            logger.warning(
                f"dictWriter.write(): argument 'mode' has invalid value '{mode}'. "
                "Used default mode 'w' instead as fallback."
            )

        # Determine target file name
        if target_file is None:
            if isinstance(source_dict, SDict) and source_dict.source_file:
                target_file = create_target_file_name(source_dict.source_file)
            else:
                logger.error("DictWriter.write(): parameter target_file is missing. No file written.")
                return

        # Make sure target_file argument is of type Path. If not, cast it to Path type.
        target_file = target_file if isinstance(target_file, Path) else Path(target_file)

        # Create formatter
        # If a formatter has been passed to write(), use that.
        # Otherwise choose the parser depending on target_file.
        formatter = formatter or Formatter.get_formatter(target_file)

        # Before writing the dict, doublecheck once again that all of its elements are correctly typed.
        parser = NativeParser()
        parser.parse_values(source_dict)

        # If mode is set to 'a' (append) and target_file exists:
        # Read the existing file and merge the new dict into the existing.
        if mode == "a" and target_file.exists():
            logger.debug(
                f"DictWriter.write(): append mode: Read existing target file {target_file} and merge dict\n"
                f"{source_dict}\n"
                "into it."
            )
            from dictIO import DictReader

            existing_dict = DictReader.read(target_file, order=order)
            existing_dict.merge(source_dict)
            source_dict = existing_dict

        # Order the dict
        if order:
            if isinstance(source_dict, SDict):
                source_dict.order_keys()
            else:
                source_dict = order_keys(source_dict)

        # Create formatted string
        string = formatter.to_string(source_dict)

        # Save formatted string to target_file
        target_file.parent.mkdir(parents=True, exist_ok=True)
        with target_file.open(mode="w") as f:
            _ = f.write(string)

        return


def create_target_file_name(
    source_file: str | os.PathLike[str],
    prefix: str | None = None,
    scope: MutableSequence[TKey] | None = None,
    output: str | None = None,
) -> Path:
    """Create a well defined target file name.

    Helper function to create a target file name based on the source file name,
    a prefix, a scope and an output format.

    Parameters
    ----------
    source_file : Union[str, os.PathLike[str]]
        source dict file
    prefix : Union[str, None], optional
        prefix to be used, by default None
    scope : Union[MutableSequence[str], None], optional
        scope to be reflected in the target file name, by default None
    output : Union[str, None], optional
        format of the target dict file. Choices are 'cpp', 'foam', 'xml' and 'json', by default None

    Returns
    -------
    Path
        target dict file name
    """
    # Make sure source_file argument is of type Path. If not, cast it to Path type.
    source_file = source_file if isinstance(source_file, Path) else Path(source_file)

    # Split source_file into file name and file ending, if existing
    file_name = source_file.stem
    file_ending = source_file.suffix

    # In case source_file is of pattern parsed.source_file (or, generically: prefix.source_file),
    # and has NO file ending, the stem/suffix approach doesn't work:
    # pathlib will interpret 'parsed' as file name (stem) and '.file_name' as file ending (suffix)
    # Let's catch that case and use a workaround to correct it:
    if str(source_file.stem) in ["parsed", prefix]:
        file_name = source_file.stem + source_file.suffix
        file_ending = ""

    # File name shall contain the scope the parsed dict had been reduced to
    if scope:
        _scope: list[str] = [str(key) for key in scope]
        scope_suffix = "_" + "_".join(_scope)
        file_name += scope_suffix

    # Prepend prefix, but make sure it is contained in the final filename max once.
    if prefix:
        prefix = prefix.removesuffix(".")  # remove trailing '.', if existing
        prefix = f"{prefix}."
        file_name = prefix + re.sub(pattern=f"^{prefix}", repl="", string=file_name)

    # If an output format is specified: Set file ending to match the output format
    if output:
        # limit to formats that are supported by DictWriter
        if output not in ["", "cpp", "foam", "json", "xml"]:
            output = "cpp"
        file_ending = "" if output == "cpp" else f".{output}"

    # Add file ending again
    file_name += file_ending

    # Compose Path object
    file_path = Path(source_file.parent, file_name)

    return file_path



"""Utility functions for working with dictionaries."""

import re
from _collections_abc import MutableMapping, MutableSequence
from collections.abc import Sequence
from copy import copy
from typing import Any

from dictIO.types import M, TGlobalKey, TKey, TValue


def order_keys(arg: M) -> M:
    """alpha-numeric sorting of keys, recursively.

    Parameters
    ----------
    arg : _MT
        MutableMapping, the keys of which shall be sorted.

    Returns
    -------
    _MT
        the passed in MutableMapping, with keys sorted. The same instance is returned.
    """
    sorted_dict: dict[TKey, Any] = dict(sorted(arg.items(), key=lambda x: (isinstance(x[0], str), x[0])))
    for key, value in copy(sorted_dict).items():
        if isinstance(value, MutableMapping):
            sorted_dict[key] = order_keys(value)  # Recursion
    arg.clear()
    arg.update(sorted_dict)
    return arg


def find_global_key(
    arg: MutableMapping[TKey, TValue] | MutableSequence[TValue],
    query: str = "",
) -> list[TGlobalKey] | None:
    """Return the global key thread to the first key the value of which matches the passed in query.

    Parameters
    ----------
    arg : Union[MutableMapping[TKey, TValue], MutableSequence[TValue]]
        dict to search in for the queried value
    query : str, optional
        query string for the value to search for, by default ''

    Returns
    -------
    Union[List, None]
        global key thread to the first key the value of which matches the passed in query, if found. Otherwise None.
    """
    global_key: list[TGlobalKey] = []
    if isinstance(arg, MutableMapping):  # dict
        for key, value in sorted(arg.items()):
            if isinstance(value, MutableMapping | MutableSequence):
                if next_level_key := find_global_key(arg=value, query=query):
                    global_key.append(key)
                    global_key.extend(next_level_key)
                    break
            elif re.search(query, str(value)):
                global_key.append(key)
                break
    else:  # list
        for index, value in enumerate(arg):
            if isinstance(value, MutableMapping | MutableSequence):
                if next_level_key := find_global_key(arg=value, query=query):
                    global_key.append(index)
                    global_key.extend(next_level_key)
                    break
            elif re.search(query, str(value)):
                global_key.append(index)
                break

    return global_key or None


def set_global_key(
    arg: MutableMapping[TKey, TValue],
    global_key: Sequence[TGlobalKey],
    value: TValue | None = None,
) -> None:
    """Set the value for the passed in global key.

    Parameters
    ----------
    arg : MutableMapping[TKey, TValue]
        dict the target key in which shall be set
    global_key : MutableSequence[TGlobalKey]
        list of keys defining the global key thread to the target key (such as returned by method find_global_key())
    value : TValue, optional
        value the target key shall be set to, by default None
    """
    if not global_key:
        return

    last_node: MutableMapping[TKey, TValue] | MutableSequence[TValue]
    next_node: MutableMapping[TKey, TValue] | MutableSequence[TValue] | TValue
    target_node: MutableMapping[TKey, TValue] | MutableSequence[TValue]
    remaining_keys: list[TGlobalKey]
    next_key: TGlobalKey
    target_key: TGlobalKey

    last_node = arg
    next_node = None
    remaining_keys = list(global_key)
    ii: int = 0
    while len(remaining_keys) > 1:
        # as long as we didn't arrive at the last node (the one that contains the target key)..
        next_key = remaining_keys[0]
        # ..walk one level further down
        if isinstance(last_node, MutableSequence):
            if not isinstance(next_key, int):
                raise KeyError(f"KeyError: {global_key} not found in {arg}")
            next_node = last_node[int(next_key)]
        else:
            next_node = last_node[next_key]
        if not isinstance(next_node, MutableMapping | MutableSequence):
            raise KeyError(f"KeyError: {global_key} not found in {arg}")
        last_node = next_node
        remaining_keys = remaining_keys[1:]
        ii += 1
        if ii == 10:  # noqa: PLR2004
            raise RecursionError(
                "RecursionError: Maximum recursion depth exceeded in set_global_key()."
                "set_global_key() is not designed to handle more than 10 levels of nested keys."
            )

    # Only one key left in remaining_keys, which is the key of the target node
    target_node = last_node
    target_key = remaining_keys[0]
    # Set the target key in target node to the passed in value
    if isinstance(target_node, MutableSequence):
        if not isinstance(target_key, int):
            raise KeyError(f"KeyError: {global_key} not found in {arg}")
        target_node[int(target_key)] = value
    else:
        target_node[target_key] = value

    return


def global_key_exists(
    dict_in: MutableMapping[TKey, TValue],
    global_key: MutableSequence[TKey],
) -> bool:
    """Check whether the specified global key exists in the passed in dict.

    Parameters
    ----------
    dict_in : MutableMapping[TKey, TValue]
        dict to check for existence of the specified global key
    global_key : MutableSequence[TKey]
        global key the existence of which is checked in the passed in dict

    Returns
    -------
    bool
        True if the specified global key exists, otherwise False
    """
    _last_branch: MutableMapping[TKey, TValue] = dict_in
    _next_branch: MutableMapping[TKey, TValue] | TValue
    try:
        for key in global_key:
            _next_branch = _last_branch[key]
            if not isinstance(_next_branch, MutableMapping):
                return False
            _last_branch = _next_branch
    except KeyError:
        return False
    return True



"""DictReader class for reading dictionaries in dictIO native file format, as well as JSON and XML."""

import contextlib
import logging
import os
import re
import sys
from collections.abc import MutableMapping, MutableSequence
from copy import deepcopy

# Import from math all functions we want to allow inside expressions in a dict.
# This is a bit ugly, but necessary to enable evaluation of parsed expressions with the help of eval().
from math import (  # noqa: F401
    acos,
    asin,
    atan,
    atan2,
    cos,
    e,
    exp,
    log,
    log10,
    pi,
    pow,
    sin,
    sqrt,
    tan,
)
from pathlib import Path
from typing import cast

from numpy import (  # noqa: F401
    array,
    diag,
    eye,
    mean,
    ndarray,
    ones,
    std,
    zeros,
)

from dictIO import Parser, SDict
from dictIO.types import TKey, TValue
from dictIO.utils.counter import DejaVue

__ALL__ = ["DictReader"]

logger = logging.getLogger(__name__)


class DictReader:
    """Reader for dictionaries in dictIO native file format, as well as JSON and XML."""

    def __init__(self) -> None:
        return

    @staticmethod
    def read(
        source_file: str | os.PathLike[str],
        *,
        includes: bool = True,
        order: bool = False,
        comments: bool = True,
        scope: MutableSequence[TKey] | None = None,
        parser: Parser | None = None,
    ) -> SDict[TKey, TValue]:
        """Read a dictionary file in dictIO native file format, as well as JSON and XML.

        Reads a dict file, parses it and transforms its content into a dictIO dict object (SDict).
        Following file formats are supported and interpreted through source_file's file ending:
        no file ending   ->   dictIO native dict file
        '.cpp'           ->   dictIO native dict file
        '.foam'          ->   Foam dictionary file
        '.json'          ->   Json dictionary file
        '.xml'           ->   XML file
        Return type is in all cases SDict

        Parameters
        ----------
        source_file : Union[str, os.PathLike[str]]
            dict file to be read
        includes : bool, optional
            merge sub-dicts being referenced through #include directives, by default True
        order : bool, optional
            sort the read dict, by default False
        comments : bool, optional
            reads comments from source file, by default True
        scope : MutableSequence[str], optional
            scope the dict will be reduced to after reading, by default None
        parser : Parser, optional
            Parser object to be used, by default None

        Returns
        -------
        SDict
            the read dict

        Raises
        ------
        FileNotFoundError
            if source_file does not exist
        """
        # Make sure source_file argument is of type Path. If not, cast it to Path type.
        source_file = source_file if isinstance(source_file, Path) else Path(source_file)
        if not source_file.exists():
            logger.error(f"source_file not found: {source_file}")
            raise FileNotFoundError(source_file)

        # Create parser
        # If a parser has been passed to read(), use that.
        # Otherwise choose the parser depending on source_file.
        # source_file = Path.joinpath(Path.cwd(), source_file)  # noqa: ERA001
        parser = parser or Parser.get_parser(source_file)

        # Parse the dict file and transform it into a SDict
        parsed_dict = parser.parse_file(source_file, comments=comments)

        # Merge dict files included through #include directives, if not actively refrained through opts
        if includes:
            DictReader._merge_includes(parent_dict=parsed_dict, comments=comments)

        # Evaluate and insert back expressions
        DictReader._eval_expressions(dict_in=parsed_dict)

        # Reduce scope of the dict if requested through opts
        if scope:
            # We need here safety hook when specified scope is not found: simply stop (not continue with whole content!)
            if parsed_dict.global_key_exists(global_key=scope):
                parsed_dict.reduce_scope(scope)
            else:
                logger.error(f"scope {scope} does not exist in dictionary")
                sys.exit(1)

        # Order the dict, if not actively refrained through opts
        if order:
            parsed_dict.order_keys()

        # Remove includes from the parsed dictionary if requested through opts
        # @TODO: Really necessary?
        # generally a good idea to have a switch suppressing #include ... in output
        # gives the option to disable include (foam required)
        # also to consider: is it really neccessary to have the included dict merged in to current dict.data?
        # if we could avoid that we get a more readable structure, even after some farn operations
        if not includes:
            DictReader._remove_include_keys(parsed_dict)

        return parsed_dict

    @staticmethod
    def _merge_includes(
        parent_dict: SDict[TKey, TValue],
        *,
        comments: bool = True,
    ) -> None:
        """Parse and merge any (child) dicts that are referenced in the dict file through #include directives."""
        # Create dejavue string watchdog
        djv = DejaVue()
        djv.reset()

        # Inner function: Merge all includes, recursively
        def _merge_includes_recursive(parent_dict: SDict[TKey, TValue]) -> SDict[TKey, TValue]:
            # empty dict to merge in temporarily, avoiding dict-has-change-error inside the for loop
            temp_dict: SDict[TKey, TValue] = SDict()

            # loop over all possible includes
            for _, _, path in parent_dict.includes.values():
                prove_recursive_include = djv(path.name)

                if prove_recursive_include is True:
                    call_chain = "->".join(list(djv.strings))
                    logger.warning(
                        f"Recursive include detected. Merging of {call_chain} into {parent_dict.name} aborted."
                    )
                elif not path.exists():
                    logger.warning(f"included dict not found. Merging of {path} aborted.")
                else:
                    parser = Parser.get_parser(source_file=path)
                    included_dict = parser.parse_file(source_file=path, target_dict=None, comments=comments)

                    # recursion in case the i-th include also has includes
                    if len(included_dict.includes) != 0:
                        nested_included_dict = _merge_includes_recursive(included_dict)
                        # merge second level
                        temp_dict.merge(nested_included_dict)

                    # merge first level
                    temp_dict.merge(included_dict)

            # merge all in for loop
            parent_dict.merge(temp_dict)

            return parent_dict

        # Call inner funtion to merge all includes, recursively
        parent_dict.merge(_merge_includes_recursive(parent_dict))

        return

    @staticmethod
    def _resolve_reference(
        reference: str,
        variables: MutableMapping[str, TValue],
    ) -> TValue:
        # resolves a single reference
        value: TValue = None
        try:
            # extract indices, ugly version, nice version is re.sub with a positive lookahead
            indexing = re.findall(pattern=r"\[.+\]$", string=reference)[0]
        except Exception:  # noqa: BLE001
            indexing = ""

        reference = re.sub(pattern=r"(^\$|\[.+$)", repl="", string=reference)  # remove leading $ or trailing [

        if reference in variables:
            value = variables[reference]  # singular value or field

            ref_changed_through_recursion = False
            while re.search(
                pattern=r"\$", string=str(value)
            ):  # resolve nested references, if existing, through recursion
                reference = str(value)
                ref_changed_through_recursion = True
                value = DictReader._resolve_reference(reference=reference, variables=variables)  # recursion
            if ref_changed_through_recursion:
                reference = re.sub(pattern=r"(^\$|\[.+$)", repl="", string=reference)  # remove leading $ or trailing [
            if indexing:
                with contextlib.suppress(Exception):
                    # return the value of the referenced variable (at the specified index, if given)
                    value = eval(f"variables['{reference}']{indexing}")  # noqa: S307
        return value

    @staticmethod
    def _eval_expressions(dict_in: SDict[TKey, TValue]) -> None:
        # Collect all references contained in expressions
        _references: list[str] = []
        _refs: list[str]
        placeholder: str
        expression: str
        for item in dict_in.expressions.values():
            _refs = re.findall(pattern=r"\$\w[\w\[\]]*", string=item["expression"])
            _references.extend(_refs)
        # Resolve references
        variables: dict[str, TValue] = dict_in.variables
        references: dict[str, TValue] = {
            ref: DictReader._resolve_reference(
                reference=ref,
                variables=variables,
            )
            for ref in _references
        }
        references_resolved: dict[str, TValue] = {
            ref: value
            for ref, value in references.items()
            if (value is not None) and (not re.search(pattern=r"EXPRESSION|\$", string=str(value)))
        }
        references_not_resolved: list[str] = [ref for ref in references if ref not in references_resolved]

        # Iteratively try to evaluate expressions contained in the dict and then re-resolve all references
        # With every iteration, this should reduce the number of remaining, non resolved references
        references_not_resolved_old: int = len(references_not_resolved) + 1
        keep_on: bool = True
        while keep_on:
            references_not_resolved_old = len(references_not_resolved)
            expressions_copy: dict[int, dict[str, str]] = deepcopy(dict_in.expressions)
            for key, item in expressions_copy.items():
                placeholder = item["name"]
                expression = item["expression"]
                _refs = re.findall(pattern=r"\$\w[\w\[\]]*", string=expression)
                for ref in _refs:
                    if ref in references_resolved:
                        expression = re.sub(
                            pattern=f"{re.escape(ref)}",
                            repl=str(references_resolved[ref]),
                            string=expression,
                        )

                eval_successful: bool = False
                eval_result: TValue | None = None
                if "$" not in expression:
                    try:
                        eval_result = eval(expression)  # noqa: S307
                        eval_successful = True
                    except NameError:
                        eval_result = expression
                        eval_successful = True
                    except SyntaxError:
                        logger.warning(f'DictReader.(): evaluation of "{expression}" not yet possible')
                if eval_successful:
                    while global_key := dict_in.find_global_key(query=placeholder):
                        # Substitute the placeholder in the dict with the result of the evaluated expression
                        dict_in.set_global_key(global_key, value=eval_result)
                    del dict_in.expressions[key]
                else:
                    # update the item in dict.expressions with the (at least partly) resolved expression
                    dict_in.expressions[key]["expression"] = expression

            # At the end of each iteration, re-resolve all references based on the now updated variables table of dict
            _references = []
            for item in dict_in.expressions.values():
                _refs = re.findall(pattern=r"\$\w[\w\[\]]*", string=item["expression"])
                _references.extend(_refs)
            variables = dict_in.variables
            references = {
                ref: DictReader._resolve_reference(
                    reference=ref,
                    variables=variables,
                )
                for ref in _references
            }
            references_resolved = {
                ref: value
                for ref, value in references.items()
                if (value is not None)
                and (
                    not re.search(
                        pattern=r"EXPRESSION|\$",
                        string=str(value),
                    )
                )
            }
            references_not_resolved = [ref for ref in references if ref not in references_resolved]

            keep_on = len(references_not_resolved) < references_not_resolved_old

        # For expressions that could NOT successfully be evaluated, even after iteration:
        # Back insert the expression string into the dict
        for item in dict_in.expressions.values():
            placeholder = item["name"]
            expression = item["expression"]
            while global_key := dict_in.find_global_key(query=placeholder):
                # Substitute the placeholder with the original (or at least partly resolved) expression
                dict_in.set_global_key(global_key, value=expression)
        dict_in.expressions.clear()

        return

    @staticmethod
    def _remove_comment_keys(data: MutableMapping[TKey, TValue]) -> MutableMapping[TKey, TValue]:
        """Remove comments from data structure for read function call from other programs."""
        remove = "[A-Z]+COMMENT[0-9;]+"

        with contextlib.suppress(Exception):
            for key in list(data.keys()):  # work on a copy of the keys
                if isinstance(data[key], MutableMapping):
                    sub_dict = cast(MutableMapping[TKey, TValue], data[key])
                    data.update({key: DictReader._remove_comment_keys(sub_dict)})  # recursion
                elif re.search(pattern=remove, string=str(key)):
                    _ = data.pop(key)
        return data

    @staticmethod
    def _remove_include_keys(data: MutableMapping[TKey, TValue]) -> None:
        """Remove includes from data structure for read function call from other programs."""
        remove = "INCLUDE[0-9;]+"

        with contextlib.suppress(Exception):
            for key in list(data.keys()):  # work on a copy of the keys
                if type(key) is str and re.search(pattern=remove, string=key):
                    _ = data.pop(key)
        return
